"""Client for interacting with Gemini AI."""

import json
import os
import re
from typing import Any, Dict, List, Optional, Tuple

import google.generativeai as genai

from .rate_limiter import RateLimiter


class GeminiClient:
    """Client for interacting with Gemini AI."""
    
    def __init__(self, verbose: bool = False, temperature: float = 0.7):
        """Initialize the Gemini client.
        
        Args:
            verbose: Whether to show verbose output
            temperature: Temperature for generation (0.0 to 1.0)
        """
        self.verbose = verbose
        
        # Configure Gemini with API key
        api_key = os.getenv("GEMINI_API_KEY")
        if not api_key:
            raise ValueError("GEMINI_API_KEY environment variable not set")
            
        genai.configure(api_key=api_key)
        
        self.model = genai.GenerativeModel(
            model_name="gemini-2.0-flash-exp",
            generation_config=genai.types.GenerationConfig(
                temperature=temperature
            )
        )
        
    async def _generate_with_retry(self, prompt: str) -> str:
        """Generate text with retries and rate limiting."""
        async def _generate():
            response = await self.model.generate_content_async(
                prompt
            )
            return response
            
        try:
            result = await RateLimiter(
                base_delay=0.5,  # 0.5 second between requests
                max_delay=32.0,  # Max 32 second delay
                max_retries=6    # Up to 6 retries
            ).execute(_generate)
            return result
        except Exception as e:
            if self.verbose:
                print(f"\033[38;5;209mError generating response: {str(e)}\033[0m\n")  # Light orange color
            raise
    
    async def _generate_and_parse_json(self, prompt: str) -> Optional[Dict]:
        """Generate a response and parse it as JSON.
        
        Args:
            prompt: Prompt to send to the model
            
        Returns:
            Parsed JSON response or None if failed
        """
        try:
            response = await self._generate_with_retry(prompt)
            
            # Handle multi-part responses
            if hasattr(response, 'parts'):
                text = '\n'.join(part.text for part in response.parts)
            else:
                text = response.text
                
            if self.verbose:
                print(f"\033[38;5;114mGemini Response:\n{text}\033[0m\n")  # Medium green color
            
            # Try to find JSON in the response
            try:
                # First check for code blocks
                code_block_match = re.search(r'```(?:json)?\s*([\s\S]*?)\s*```', text)
                if code_block_match:
                    try:
                        return json.loads(code_block_match.group(1))
                    except:
                        pass
                
                # Then try to parse the entire text
                try:
                    return json.loads(text.strip())
                except:
                    pass
                
                # Finally try to find JSON-like content
                text = text.strip()
                if text.startswith('['):
                    # Find matching closing bracket
                    count = 0
                    for i, char in enumerate(text):
                        if char == '[':
                            count += 1
                        elif char == ']':
                            count -= 1
                            if count == 0:
                                try:
                                    return json.loads(text[:i+1])
                                except:
                                    pass
                                break
                
                # Try to find object matches
                matches = re.findall(r'\{[^}]+\}', text)
                if matches:
                    return json.loads(matches[0])
                
                if self.verbose:
                    print(f"\033[38;5;209mNo JSON found in response\033[0m\n")  # Light orange color
                return None
                    
            except Exception as e:
                if self.verbose:
                    print(f"\033[38;5;209mError parsing JSON: {str(e)}\033[0m\n")  # Light orange color
                return None
            
        except Exception as e:
            if self.verbose:
                print(f"\033[38;5;209mError generating response: {str(e)}\033[0m\n")  # Light orange color
            return None
            
    async def parse_query(
        self,
        query: str
    ) -> Tuple[str, List[str], Optional[str]]:
        """Parse a natural language query into structured components."""
        prompt = f"""Parse this query into structured components:
{query}

Output format:
entity_type: <type of entity being searched for>
attributes: <list of attributes to extract, comma separated>
search_space: <optional geographic or domain-specific scope>

Rules:
1. entity_type should be a singular noun (e.g. "company" not "companies")
2. attributes should be normalized:
   - "name" and "names" -> "name"
   - "website" and "websites" -> "website"
   - "phone" and "phones" -> "phone"
   - "email" and "emails" -> "email"
3. search_space should capture any geographic or domain constraints
4. if no search_space is specified, output "none"

Example 1:
Query: "find tech companies in california with their websites and phone numbers"
Output:
entity_type: company
attributes: website,phone
search_space: california

Example 2:
Query: "list universities with email addresses"
Output:
entity_type: university
attributes: email
search_space: none"""

        try:
            response = await self._generate_with_retry(prompt)
            
            # Handle multi-part responses
            if hasattr(response, 'parts'):
                text = '\n'.join(part.text for part in response.parts)
            else:
                text = response.text
                
            if self.verbose:
                print(f"\033[38;5;114mGemini Response:\n{text}\033[0m\n")  # Medium green color
            
            # Parse response
            lines = text.strip().split('\n')
            entity_type = None
            attributes = []
            search_space = None
            
            for line in lines:
                if ':' not in line:
                    continue
                    
                key, value = line.split(':', 1)
                key = key.strip().lower()
                value = value.strip().lower()
                
                if key == 'entity_type':
                    entity_type = value
                elif key == 'attributes':
                    # Split and clean attributes
                    attrs = [a.strip() for a in value.split(',')]
                    # Normalize attribute names
                    normalized = []
                    for attr in attrs:
                        if attr in ['name', 'names']:
                            continue  # Skip 'name' since it's always included
                        elif attr in ['website', 'websites']:
                            normalized.append('website')
                        elif attr in ['phone', 'phones']:
                            normalized.append('phone')
                        elif attr in ['email', 'emails']:
                            normalized.append('email')
                        elif attr in ['director', 'directors']:
                            normalized.append('director')
                        else:
                            normalized.append(attr)
                    attributes = normalized
                elif key == 'search_space' and value != 'none':
                    search_space = value
            
            if not entity_type:
                raise ValueError("No entity type found in response")
            
            return entity_type, attributes, search_space
            
        except Exception as e:
            if self.verbose:
                print(f"\033[38;5;209mError parsing query: {str(e)}\033[0m\n")  # Light orange color
            raise
    
    async def enumerate_search_space(self, search_space: str) -> List[str]:
        """Enumerate items in a search space.
        
        Args:
            search_space: Description of the search space
            
        Returns:
            List of items in the search space
        """
        prompt = f"""Enumerate all items in this search space:
"{search_space}"

Return a JSON array of strings. For geographic areas, use official names.
For example, for "New England states" return:
[
    "Maine",
    "New Hampshire",
    "Vermont",
    "Massachusetts",
    "Rhode Island",
    "Connecticut"
]"""
        
        return await self._generate_and_parse_json(prompt)
    
    async def extract_attributes(self, text: str, url: str, entity_type: str, attributes: List[str]) -> Optional[Dict[str, str]]:
        """Extract requested attributes from text.
        
        Args:
            text: Text to extract from
            url: URL the text came from
            entity_type: Type of entity to extract
            attributes: List of attributes to extract
            
        Returns:
            Dictionary of extracted attributes or None if failed
        """
        prompt = f"""Extract information about this {entity_type} from the following text:

Text: {text}
URL: {url}

Extract these attributes:
{', '.join(attributes)}

Rules:
1. Only include attributes that are explicitly mentioned in the text
2. Do not guess or infer values
3. For addresses, include as much detail as given (street, city, state, zip)
4. Return values exactly as they appear in text
5. Omit attributes if no value is found

Return as JSON with the attribute names as keys.
Example:
{{
    "name": "LA Fitness",
    "address": "123 Main St, Tampa, FL 33601"
}}"""

        try:
            response = await self._generate_with_retry(prompt)
            
            if hasattr(response, 'parts'):
                text = '\n'.join(part.text for part in response.parts)
            else:
                text = response.text
                
            if self.verbose:
                print(f"\033[38;5;114mGemini Response:\n{text}\033[0m\n")  # Medium green color
            
            # Try to find JSON in the response
            try:
                # First check for code blocks
                code_block_match = re.search(r'```(?:json)?\s*([\s\S]*?)\s*```', text)
                if code_block_match:
                    try:
                        return json.loads(code_block_match.group(1))
                    except:
                        pass
                
                # Then try to parse the entire text
                try:
                    return json.loads(text.strip())
                except:
                    pass
                
                # Finally try to find JSON-like content
                text = text.strip()
                if text.startswith('{'):
                    # Find matching closing brace
                    count = 0
                    for i, char in enumerate(text):
                        if char == '{':
                            count += 1
                        elif char == '}':
                            count -= 1
                            if count == 0:
                                try:
                                    return json.loads(text[:i+1])
                                except:
                                    pass
                                break
                
                if self.verbose:
                    print(f"\033[38;5;209mNo JSON found in response\033[0m\n")  # Light orange color
                return None
                    
            except Exception as e:
                if self.verbose:
                    print(f"\033[38;5;209mError parsing JSON: {str(e)}\033[0m\n")  # Light orange color
                return None
            
        except Exception as e:
            if self.verbose:
                print(f"\033[38;5;209mError extracting attributes: {str(e)}\033[0m\n")  # Light orange color
            return None
    
    async def parse_search_result(
        self,
        text: str,
        entity_type: str,
        attributes: List[str]
    ) -> Optional[Dict[str, str]]:
        """Parse a search result to extract requested attributes.
        
        Args:
            text: Text to parse (title + snippet)
            entity_type: Type of entity to extract
            attributes: List of attributes to extract
            
        Returns:
            Dictionary of extracted attributes or None if no match
        """
        return await self.extract_attributes(text, "", entity_type, attributes)
